#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Helper script to launch AiiDAlab on a local workstation.

.. currentmodule:: aiidalab_launch
.. moduleauthor:: Carl Simon Adorf <simon.adorf@epfl.ch>
"""

from .version import __version__

__all__ = [
    "__version__",
]
